unit AccessibilityUtils;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, VAUtils;

const
  ST_DISABLED : string = '.  Disabled field.  Not applicable.';

procedure ConfigDisabledEditBoxForScreenReader(Edit: TEdit);
function  UpdateEditBoxesForScreenReader(Form: TForm): boolean;
procedure SetCursorToStart(Sender: TObject);
function OddDateFormatToDelphiDate(DateTimeST: String): TDateTime;

implementation

{==============================================================================
  ConfigDisabledEditBoxForScreenReader
    ConfigDisabledEditBoxForScreenReader is a method that will set the properties
      of a TEdit so that it appears to be disabled, but will still be accessible
      to a screen reader

  Parameters:
    Edit: Edit is the Edit on which the properties will be modified

  Modification History:
    CodeCR116 - 8/2010 Method created -MER
 ==============================================================================}
procedure ConfigDisabledEditBoxForScreenReader(Edit: TEdit);
begin
  Edit.Enabled := TRUE;
  Edit.ReadOnly := TRUE;
  Edit.Font.Color := clGray;
  Edit.AutoSelect := FALSE;
end;

{==============================================================================
  UpdateEditBoxesForScreenReader
    UpdateEditBoxesForScreenReader is a method that will correct a 508 issue
      with edit boxes and screen readers.  In some places, CAPRI will set an
      Edit Box to disabled to allow a user to see information that they cannot
      modify.  A disabled component does not have an active tab stop, so such
      editBoxes would not be effectively visible to a screen reader.  In order
      to allow such controls to have a tab stop, appear to be disabled, and
      still be not editable, we can set readOnly and change the text color to
      clGrayText
    UpdateEditBoxesForScreenReader first checks if a screen reader is active.  If
      one is active, we iterate through all the components on the form and
      modify all TEdits (and descendants of TEdit) that are both visible and
      disabled

  Parameters:
    Form: Form is the TForm on which the edit fields will be modified

  Return Value:
    UpdateEditBoxesForScreenReader returns a boolean.  It will return true if
      a screen reader is active, otherwise it will return false

  Modification History:
    CodeCR116 - 8/2010 Method created -MER
 ==============================================================================}
function UpdateEditBoxesForScreenReader(Form: TForm): boolean;
var
  ndx: integer;
begin
  Result := FALSE;
  if ScreenReaderActive then begin
    for ndx := 0 to (Form.ComponentCount - 1) do begin
      if (Form.Components[ndx] is TEdit) and (TEdit(Form.Components[ndx]).Visible) and (TEdit(Form.Components[ndx]).Enabled = FALSE)
      then
        ConfigDisabledEditBoxForScreenReader(TEdit(Form.Components[ndx]));
    end;
    Result := TRUE;  // indicate screen reader is active
  end;
end;

{==============================================================================
  SetCursorToStart
    SetCursorToStart is a method that will cause the cursor to be set to the start
      of any TMemo that is passed to it.  At the time this was created, only
      TMemos are handled, but the parameter is a more genric TObject in case
      it's use is expanded in the future

    Parameters:
      Sender: Sender is the TObject on which the cursor location will be
        modified.

  Modification History:
    CodeCR116 - 8/2010 Method created -MER
 ==============================================================================}
procedure SetCursorToStart(Sender: TObject);
begin
  if Sender is TMemo then begin
    with Sender as TMemo do begin
      selStart := 0;
    end;
  end;
end;

{==============================================================================
  OddDateFormatToDelphiDate
    OddDateFormatToDelphiDate is a function that will return a TDateTime from
      a format of either MMM dd, yyyy or MMM dd,yyyy.  The only difference in
      the previous formats is the inclusion of a space after the comma.
      This function was made as a modification of ConvertCalendarDateToDelphiDate
      due to the issue where the displayed format and RPC validated format were
      different (and could not be used for both).  If an exception is raised,
      this function does not trap the exception to allow it to be handled at a
      higher level

  Parameters:
    DateTimeST: DateTimeST is a string that should have a format of either
      MMM dd, yyyy or MMM dd,yyyy

  Return value:
    OddDateFormatToDelphiDate returns a TDateTime representing the date passed
      in DateTimeST

  Modification History:
    CodeCR116 - 8/2010 Method created -MER
 ==============================================================================}
function OddDateFormatToDelphiDate(DateTimeST: String): TDateTime;
var
  Month, Day, Year: String;
begin
  Month := Uppercase(Piece(DateTimeST, ' ', 1));
  Day := Piece(Piece(DateTimeST, ' ', 2), ',', 1);
  Year := TrimLeft(Piece(DateTimeST, ',', 2));
  if Pos(Month, 'JAN') = 1 then
    Month := '01';
  if Pos(Month, 'FEB') = 1 then
    Month := '02';
  if Pos(Month, 'MAR') = 1 then
    Month := '03';
  if Pos(Month, 'APR') = 1 then
    Month := '04';
  if Pos(Month, 'MAY') = 1 then
    Month := '05';
  if Pos(Month, 'JUN') = 1 then
    Month := '06';
  if Pos(Month, 'JUL') = 1 then
    Month := '07';
  if Pos(Month, 'AUG') = 1 then
    Month := '08';
  if Pos(Month, 'SEP') = 1 then
    Month := '09';
  if Pos(Month, 'OCT') = 1 then
    Month := '10';
  if Pos(Month, 'NOV') = 1 then
    Month := '11';
  if Pos(Month, 'DEC') = 1 then
    Month := '12';
  if Length(Day) = 1 then
    Day := '0' + Day;
  Result := StrToDate(Month + '/' + Day + '/' + Year);
end; // OddDateFormatToDelphiDate

end.
