unit virtualVAWebServiceClient;

(*
   The TvvaWSC class provides the interface between CAPRI and the Virtual VA
   web service client DLL.  The primary use is to accept document text in
   the form of a stringlist, format the report using Rave Reports, and
   generate a PDF file, which is then transmitted to the Virtual VA web
   service utilizing the VACAPRIVVA.dll web services client.  An event
   trigger is provided at specific action points for report format,
   transmission, and file deletion to allow a consuming application to
   listen for these status changes.

   The class contains methods for passing in an existing PDF file name
   for transmission and for selecting an existing PDF file for transmission
   using a File Open dialog, but are not utilized at this time.

   After instantiation, the consuming application must populate all fields
   in the TvvaPatient, TvvaDoc, and TvvaReport records.  This data is used
   in the report creation and to populate the Virtual VA web service required
   fields.

   NOTE:  The Claim Number indexing value is called eFolder in the
          requirements, but is called claimnumber in the web service
          property definition.  The requirements specify that this
          value be populated with the veteran's Social Security Number
          and not the Claim number.

   ADDITIONAL NOTES:  The design of Virtual VA utilizes QuickReport
   document content that is fed into Rave Reports with PDF conversion
   provided by the Rave Reports TRvRenderPDF class.
   An alternate design using the Synopse framework for PDF conversion that
   accepts QuickReport data directly was implemented as a potential
   alternative.  The Synopse design is encapsulated in the QR2PDFConverter
   unit. The overloaded CreateXmitFile method implements the PDF converter.
   The method remains in this unit for possible future utilization, but
   it and the QR2PDFConverter unit in the uses clause are commented out.
*)

interface

uses
  Classes, ComCtrls, Controls, Dialogs, Graphics, QRCtrls, QRPrntr, QuickRpt,
  {QR2PDFConverter,} RpRender, RpRenderPDF, RpDefine, RpBase, RpSystem,
  SysUtils, VAUtils, Windows, ShellAPI, dialogsCAPRI;

type
  TvvaAction = (actValidate, actCreate, actSend, actDelete, actComplete);
  TvvaStatus = (stFailed, stSuccess, stPending, stProcessing);

  TNotifyStatus = procedure(Sender: TObject; Action: TvvaAction; Status: TvvaStatus) of object;

  TvvaPatient = record
    FullName: String;
    SSN: String;
  end;

  TvvaDoc = record  //Virtual VA web service parameters
    URL: String;      //web services URL
    SecToken: String;  //web services security
    User: String;
    DocType: Integer;  //CAPRI
    DocCategory: Integer;  //Medical Records
    Subject: String;  //Clinical Documents
    Source: String;  //CAPRIVISTA
    SourceComment: String;  //Station name and number
    OutputFilePath: String;
    OutputFileName: String;
  end;

  TvvaReport = record   //Report generation parameters
    ReportText: TStringList;
    FooterDesc: String;
    FontSize: Integer;
    FontName: String;
    SystemName: String;
    Division: String;
    PrintDate: String;
    Patient: TvvaPatient;
    VVADoc: TvvaDoc;
  end;

  TvvaWSC = class(TObject)
    RaveSystem: TRvSystem;
    RaveRenderPDF: TRvRenderPDF;
    procedure RavePrint(Sender: TObject);
    procedure RavePrintFooter(Sender: TObject);
    procedure RavePrintHeader(Sender: TObject);
    procedure RaveBeforePrint(Sender: TObject);
  private
    { Private declarations }
    FOnUpdateStatus : TNotifyStatus;
    FIsRunning: Boolean;
    FDeleteFile: Boolean;
    VVAOpInProcess: Boolean;
    function GetXmitFile(var vvaReport: TvvaReport): Boolean;
    function XmitFile(vvaReport: TvvaReport): String;
    function ReplaceTabs(LineWithTabs: String; tabSize: Integer): String;
    procedure PrintLine(ReportComponent: TBaseReport; ReportLine: String);
    function IsFileExists(): Boolean;
    function GetFileName(const FileDescription: String): String;
    // function GetTempDirectory: String;
    function IsVVAReportValid(var vvaReport: TvvaReport; var ErrorText: String): Boolean;
    procedure FormatIndexTerms(var vvaReport: TvvaReport);
    function GetFileSize(const FileName: String): LongInt;
  protected
    function makeUniqueFileName(const FileName: String; const SSN: String): String;  // CodeCR353 LMS 2013-01-29 2013-01-31
    procedure DoUpdateStatus(Action: TvvaAction; Status : TvvaStatus); virtual;
  public
    { Public declarations }
    vvaReport: TvvaReport;
    property OnUpdateStatus : TNotifyStatus read FOnUpdateStatus
                                       write FOnUpdateStatus;
    property IsRunning: Boolean read FIsRunning;
    property IsDeleteFile: Boolean read FDeleteFile write FDeleteFile;
    function IsConfirmedToSend(aPatientName: String): Boolean;
    function SendReportToVVA: String;
    function SendFileToVVA: String;
    function GetAndSendFileToVVA: String;
    procedure LoadText(QRComponent : TQRRichText); overload;
    procedure LoadText(aStringList: TStringList); overload;
    function CreateXmitFile(var vvaReport: TvvaReport): LongInt;
//    function CreateXmitFile(aQuickRpt: TQuickRep; afooterDesc: String): LongInt; overload;
    function FormatByteSize(const bytes: LongInt): string;
    constructor Create;
    destructor Destroy; Override;
    procedure CompressPDF(UncompressedFilename, CompressedFilename: String);
  end;

  TApplicationProxy = class(TObject)  // LMS Extra work to launch threads within DLL and also have DLL processmessages for the main CAPRI message queue.
  private                             //      Using XE2's TThread implementation instead of doing it within CAPRI's 2006 code so it's using more updated OS calls.
    FProxy : TObject;                 //     There is no Application object in the DLL.
    FProxyProcessMessagesPointer : Pointer;
    FWasMultiThreaded : Boolean;
  public
    property ProxyProcessMessagesPointer : Pointer read FProxyProcessMessagesPointer write FProxyProcessMessagesPointer;
    constructor Create; overload; virtual;
    destructor Destroy; override;
  published
    procedure ProcessMessages;  // LMS Publishing to get it via RTTI and pass it to DLL. Can't rely on XE2's call to MethodAddress as TObject may have changed.
  end;

var
  vvaWSC: TvvaWSC;

function SendVVADoc(const _ClaimNum: WideString; const _Pth: WideString;
                    const _DocTitle: WideString; const _FilNme: WideString;
                    const _DocType: Integer; const _DocSubj: WideString;
                    const _DocCat: Integer; const _Src: WideString;
                    const _Usr: WideString; const _SrcCmnt: WideString;
                    const _URL: WideString;
                    const _SECUSR: WideString;
                    const _SECPWD: WideString;
                    const _SECTOKEN: WideString): ShortString; cdecl; external 'VACAPRIVVA.dll';
implementation

uses
  frmVVAConfirmDialog
  , synCommons, math   // CodeCR353 LMS 2013-01-29
  , forms, Main, CCOWRPCBrokerCAPRI    // CodeCR424 LMS 2013-05-07
  , untMiscMthds
  ;

{============================================================================
 LoadText
 This overloaded procedure converts the RTF text provided by the TQRRichText
 component to plain text and loads the results into vvaReport.ReportText.
 ============================================================================}
procedure TVVAWSC.LoadText(QRComponent: TQRRichText);
var
  ms: TMemoryStream;
  richText: TRichEdit;
begin
  ms := TMemoryStream.Create;
  richText := TRichEdit.CreateParented(HWND_MESSAGE);
  //set MaxLength to the maximum of $7FFFFFFD to allow for large document
  //sizes to be combined and converted to plain text
  richText.MaxLength := 2147483645;
  richText.PlainText := False;
  vvaReport.ReportText.Clear;
  try
    QRComponent.Lines.SaveToStream(ms);
    ms.Seek(0, soFromBeginning);
    richText.Lines.LoadFromStream(ms);
    ms.Clear;
    richText.PlainText := True;
    richText.Lines.SaveToStream(ms);
    ms.Seek(0, soFromBeginning);
    vvaReport.ReportText.LoadFromStream(ms);
  finally
    ms.Free;
    richText.Free;
  end;
end;

{============================================================================
 LoadText
 This overloaded procedure loads a passed StringList into vvaReport.ReportText.
 ============================================================================}
procedure TVVAWSC.LoadText(aStringList: TStringList);
var
  ms: TMemoryStream;
begin
  ms := TMemoryStream.Create;
  vvaReport.ReportText.Clear;
  try
    aStringList.SaveToStream(ms);
    ms.Seek(0, soFromBeginning);
    vvaReport.ReportText.LoadFromStream(ms);
  finally
    ms.Free;
  end;
end;


constructor TVVAWSC.Create;
begin
  inherited;
  VVAOpInProcess := false;

  vvaReport.ReportText := TStringList.Create;
  RaveSystem := TRvSystem.Create(nil);
  with RaveSystem do
  begin
    TitleSetup := 'Output Options';
    TitleStatus := 'Report Status';
    TitlePreview := 'Report Preview';
    SystemSetups := [ssAllowCollate, ssAllowDuplex, ssAllowDestPreview, ssAllowDestPrinter, ssAllowDestFile, ssAllowPrinterSetup, ssAllowPreviewSetup];
    SystemOptions := [soAllowPrintFromPreview, soAllowSaveFromPreview, soPreviewModal];
    SystemFiler.StatusFormat := 'Generating page %p';
    SystemPreview.ZoomFactor := 100.000000000000000000;
    SystemPrinter.MarginBottom := 0.500000000000000000;
    SystemPrinter.MarginLeft := 0.500000000000000000;
    SystemPrinter.MarginRight := 0.500000000000000000;
    SystemPrinter.MarginTop := 0.500000000000000000;
    SystemPrinter.ScaleX := 100.000000000000000000;
    SystemPrinter.ScaleY := 100.000000000000000000;
    SystemPrinter.StatusFormat := 'Printing page %p';
    SystemPrinter.Title := 'Rave Report';
    SystemPrinter.UnitsFactor := 1.000000000000000000;
  end;

  RaveRenderPDF := TRvRenderPDF.Create(nil);
  with RaveRenderPDF do
  begin
    DisplayName := 'Adobe Acrobat (PDF)';
    FileExtension := '*.pdf';
    EmbedFonts := True;
    ImageQuality := 90;
    MetafileDPI := 300;
    FontEncoding := feWinAnsiEncoding;
    DocInfo.Creator := 'Rave Reports (http://www.nevrona.com/rave)';
    DocInfo.Producer := 'Nevrona Designs';
    BufferDocument := True;
    DisableHyperlinks := False;
  end;
  FIsRunning := False;
  FDeleteFile := True;  //default to deleting the temporary PDF after transmission
end;

destructor TVVAWSC.Destroy;
begin
  FreeAndNil(RaveSystem);
  FreeAndNil(RaveRenderPDF);
  FreeAndNil(vvaReport.ReportText);
  inherited;
end;

{=========================================================================
 DoUpdateStatus
 This procedure is the event trigger to fire the Update messages
 =========================================================================}
procedure TvvaWSC.DoUpdateStatus(Action: TvvaAction; Status: TvvaStatus);
begin
  if Assigned(FOnUpdateStatus) then
    FOnUpdateStatus(Self, Action, Status);
end;

{=========================================================================
 FormatIndexTerms
 This procedure truncates the length of all index terms to specification.
 =========================================================================}
procedure TvvaWSC.FormatIndexTerms(var vvaReport: TvvaReport);
begin
  vvaReport.VVADoc.SourceComment := Copy(vvaReport.VVADoc.SourceComment, 1, 50);
  vvaReport.VVADoc.Subject := Copy(vvaReport.VVADoc.Subject, 1, 80);
  vvaReport.VVADoc.User := Copy(vvaReport.VVADoc.User, 1, 20);
  vvaReport.VVADoc.Source := Copy(vvaReport.VVADoc.Source, 1, 10);
end;

{==========================================================================
 IsConfirmedToSend
 This function displays the confirmation dialog and returns True when OK
 is selected; otherwise it returns False.
 A manually created confirmation dialog is used, because the JAWS screen
 reader does not support the standard MessageDlg call.
===========================================================================}
function TVVAWSC.IsConfirmedToSend(aPatientName: String): Boolean;
var
  frmConfirm: TfrmConfirmDlg;
begin
  frmConfirm := TfrmConfirmDlg.Create(nil);
  try
    if VAUTils.ScreenReaderActive then
      frmConfirm.edConfirm.TabStop := True
    else
      frmConfirm.edConfirm.TabStop := False;

    frmConfirm.edConfirm.Text := 'Are you sure you want to send document(s) to the VBA record of ' + aPatientName + '?';
    frmConfirm.edConfirm.Width := frmConfirm.Canvas.TextWidth(frmConfirm.edConfirm.Text) + 8;
    frmConfirm.Width := frmConfirm.edConfirm.Width + 70;
    frmConfirm.Image1.Picture.Icon.Handle := LoadIcon(0, IDI_QUESTION);

  finally
    Result := (frmConfirm.ShowModal = mrYes);
    FreeAndNil(frmConfirm);
  end;
end;

{==========================================================================
 IsVVAReportValid
 This function formats and validates the input terms
 Output:  returns True and empty ErrorText on success; Otherwise returns
          False and ErrorText reasons.
 ==========================================================================}
function TvvaWSC.IsVVAReportValid(var vvaReport: TvvaReport; var ErrorText: String): Boolean;
begin
  FormatIndexTerms(vvaReport);
  ErrorText := '';
  if (vvaReport.VVADoc.DocType <= 0) then
    ErrorText := ErrorText + #13#10 + 'Document Type missing';
  if vvaReport.Patient.SSN = '' then
    ErrorText := ErrorText + #13#10 + 'eFolder number missing';  //see NOTE at top of unit.
  if (Length(vvaReport.Patient.SSN) > 9) or
     (StrToIntDef(vvaReport.Patient.SSN, -1) = -1) then
    ErrorText := ErrorText + #13#10 + vvaReport.Patient.SSN + ' is not a valid eFolder number';
  if (vvaReport.VVADoc.DocCategory <= 0) then
    ErrorText := ErrorText + #13#10 + 'Document Category missing';
  if vvaReport.VVADoc.Subject = '' then
    ErrorText := ErrorText + #13#10 + 'Subject missing';
  if vvaReport.VVADoc.Source = '' then
    ErrorText := ErrorText + #13#10 + 'Source missing';
  if vvaReport.VVADoc.SourceComment = '' then
    ErrorText := ErrorText + #13#10 + 'Source Comment missing';
  if vvaReport.VVADoc.User = '' then
    ErrorText := ErrorText + #13#10 + 'User ID missing';

  if (ErrorText <> '') then
  begin
    Result := False;
    ErrorText := 'Validation Errors: ' + ErrorText;
  end
  else
  begin
    Result := True;
  end;
end;

{==========================================================================
 IsFileExists
 This function validates a passed filename and filepath to determine if the
 file already exists.
 ==========================================================================}
 function TVVAWSC.IsFileExists(): Boolean;
 begin
   Result := (vvaReport.vvaDoc.OutputFilePath <> '') and
             (vvaReport.vvaDoc.OutputFileName <> '') and
     FileExists(vvaReport.vvaDoc.OutputFilePath + vvaReport.vvaDoc.OutputFileName);
 end;

{==========================================================================
 GetandSendFileToVVA
 This procedure transmits an existing file to the Virtual VA web service.
 A File Open dialog is used to allow selection of an existing file.
 ==========================================================================}
function TvvaWSC.GetAndSendFileToVVA: String;
var
  ErrorMsg: String;
begin
  ErrorMsg := '';
  Result := '';
  FIsRunning := True;
  try
    if IsVVAReportValid(vvaReport, ErrorMsg) then
    begin
      if GetXmitFile(vvaReport) and IsFileExists then  //selected file valid
      begin
        DoUpdateStatus(actCreate, stSuccess);
        if IsConfirmedToSend(vvaReport.Patient.FullName) then
        begin
          DoUpdateStatus(actSend, stProcessing);
          Result := XmitFile(vvaReport);  //Transmit file
        end;
      end
      else
        DoUpdateStatus(actCreate, stFailed);
        Result := 'No file to send.';
    end
    else
      Result := ErrorMsg;
      DoUpdateStatus(actValidate, stFailed);
  finally
    DoUpdateStatus(actComplete, stSuccess);
    FIsRunning := False;
  end;
end;

{==========================================================================
 SendFileToVVA
 This procedure transmits an existing file to the Virtual VA web service.
 The file name and path must be passed and specify a valid existing file.
 ==========================================================================}
function TvvaWSC.SendFileToVVA: String;
var
  ErrorMsg: String;
begin
  ErrorMsg := '';
  FIsRunning := True;
  try
    if IsVVAReportValid(vvaReport, ErrorMsg) then
    begin  //Valid record
      if IsFileExists then  //passed filepath/filename point to a valid file
      begin
        DoUpdateStatus(actCreate, stSuccess);
        DoUpdateStatus(actSend, stProcessing);
        Result := XmitFile(vvaReport);  //Transmit file to VVA
        if FDeleteFile  and
           DeleteFile(PAnsiChar(vvaReport.vvaDoc.OutputFilePath + vvaReport.vvaDoc.OutputFileName)) and
           DeleteFile(PAnsiChar(vvaReport.vvaDoc.OutputFilePath + 'c' + vvaReport.vvaDoc.OutputFileName)) then // CodeCR559 JRL 3/13/14
          DoUpdateStatus(actDelete, stSuccess) // File deletion succeeded
      else  //File deletion failed
        DoUpdateStatus(actDelete, stFailed);
      end
      else
      begin
        DoUpdateStatus(actCreate, stFailed);
        Result := 'No file to send.';
      end;
    end  //Valid record
    else
    begin  //Invalid record
      Result := ErrorMsg;
      DoUpdateStatus(actValidate, stFailed);
    end;  //Invalid record
  finally
    DoUpdateStatus(actComplete, stSuccess);
    FIsRunning := False;
  end;
end;

{==========================================================================
 SendReportToVVA
 This procedure creates a transmit file from a pre-loaded StringList,
 transmits the file to the Virtual VA and deletes the file on completion.
 ==========================================================================}
function TVVAWSC.SendReportToVVA: String;
var
  ErrorMsg: String;
begin
  Result := '';
  FIsRunning := True;
  try
    if IsVVAReportValid(vvaReport, ErrorMsg) then
    begin  //Valid record
      //Create a formatted PDF file or retrieve existing file to transmit to VVA
      if (CreateXmitFile(vvaReport) > 0) then  //File ready to transmit
      begin
        DoUpdateStatus(actSend, stProcessing);
        Result:= XmitFile(vvaReport);   //Transmit file to VVA
      end;

      if FDeleteFile and
         DeleteFile(PAnsiChar(vvaReport.vvaDoc.OutputFilePath + vvaReport.vvaDoc.OutputFileName)) and 
         DeleteFile(PAnsiChar(vvaReport.vvaDoc.OutputFilePath + 'c' + vvaReport.vvaDoc.OutputFileName)) then //CodeCR559 JRL 03/12/14
        DoUpdateStatus(actDelete, stSuccess)  // File deletion succeeded
      else  //File deletion failed
        DoUpdateStatus(actDelete, stFailed);
    end  //Valid record
    else
    begin  //Invalid record
      Result := ErrorMsg;
      DoUpdateStatus(actValidate, stFailed);
    end;  //Invalid record
  finally
    DoUpdateStatus(actComplete, stSuccess);
    FIsRunning := False;
  end;
end;

{============================================================================
 GetXmitFile
 This function retrieves the file to transmit through a file dialog.

 Input:  vvaReport: TvvaReport - report object
 Output: function result - returns True on success;
                           otherwise returns False.
 ============================================================================}
function TVVAWSC.GetXmitFile(var vvaReport: TvvaReport): Boolean;
var
  vvaFileDialog: TOpenDialog;
begin
  vvaFileDialog := TOpenDialog.Create(nil);
  vvaFileDialog.Filter := 'Adobe PDF files|*.pdf'
                          + '|Excel files|*.xl*;*.xlsx'
                          + '|Text files|*.txt'
                          + '|Word document files|*.doc;*.docx'
                          + '|All files|*.*';
  try
    if vvaFileDialog.Execute then
    begin
      vvaReport.VVADoc.OutputFilePath := ExtractFilePath(vvaFileDialog.FileName);
      vvaReport.VVADoc.OutputFileName := vvaFileDialog.FileName;
      Result := True;
    end
    else
    begin
      Result := False;
    end;
  finally
    vvaFileDialog.Free;
  end;
end;

{============================================================================
 CreateXmitFile
 This function uses Rave Reports to format and create the output file that
 will be transmitted to the Virtual VA.

 Input:  vvaReport: TvvaReport - report object
 Output: function result - returns True on success;
                           otherwise returns False.
 CodeCR559 3/18/14 JRL - Take the existing PDF and add compression and 
                         linearization and create a second "compressed" PDF
                         for sending to VLER/DAS.
 ============================================================================}
function TVVAWSC.CreateXmitFile(var vvaReport: TvvaReport): LongInt;
const
  FILE_EXTENT = '.PDF';
begin
  //Specify filepath
  vvaReport.VVADoc.OutputFilePath := GetTempDir;  //  CodeCR496 - LMS - 2014-07-14 refactor GetTempDir to one method call

  //Specify filename
  vvaReport.VVADoc.OutputFileName := GetFileName(vvaReport.FooterDesc) + FILE_EXTENT;

  RaveSystem.OnPrint := RavePrint;
  RaveSystem.OnPrintHeader := RavePrintHeader;
  RaveSystem.OnPrintFooter := RavePrintFooter;
  RaveSystem.DefaultDest := rdFile;
  RaveSystem.DoNativeOutput := false;
  RaveSystem.RenderObject := RaveRenderPDF;
  RaveSystem.OutputFileName := vvaReport.vvaDoc.OutputFilePath + vvaReport.vvaDoc.OutputFileName; 
  RaveSystem.SystemSetups := RaveSystem.SystemSetups - [ssAllowSetup];
  RaveSystem.Execute;
  if FileExists(RaveSystem.OutputFileName) then
  begin
    DoUpdateStatus(actCreate, stSuccess);
  end
  else
  begin
    DoUpdateStatus(actCreate, stFailed);
  end;

  // make a compressed version of the PDF here to send to VLER/DAS - filename has a 'c' added at the start  
  // CodeCR559 JRL 3/17/14
  try
    CompressPDF(RaveSystem.OutputFileName,vvaReport.VVADoc.OutputFilePath + 'c' + vvaReport.VVADoc.OutputFileName); 
  except                                                                                  
    on E:Exception do                                                                     
      begin // Copy uncompressed file to compressed filename so it can be sent if there was an error compressing
        if FileExists(vvaReport.VVADoc.OutputFilePath + 'c' + vvaReport.VVADoc.OutputFileName) then
           DeleteFile(pchar(vvaReport.VVADoc.OutputFilePath + 'c' + vvaReport.VVADoc.OutputFileName));
        if FileExists(RaveSystem.OutputFileName) then
           MoveFile(pchar(RaveSystem.OutputFileName), pchar(vvaReport.VVADoc.OutputFilePath + 'c' + vvaReport.VVADoc.OutputFileName));
      end; // compress except
  end; // compress try
                                                                                     
  Result := GetFileSize(vvaReport.VVADoc.OutputFilePath + vvaReport.VVADoc.OutputFileName);  // Return file size                         
end;


{============================================================================
 CompressPDF
 This function uses an external tool, QPDF, to compress the Rave Report 
 output PDF file to a compressed and linearized format.  This file will be
 transmitted to the Virtual VA.  QPDF must finish compressing the file before
 exiting from this routine to avoid file errors in subsequent code.

 Parameters:  
    Input:  Uncompressed PDF filename 
    Output: Compressed PDF filename 

 Required files for running QPDF:
    qpdf.exe
    qpdf13.dll
    libgcc_s_dw2-1.dll
    libstdc++6.dll

 Created routine:  CodeCR559 3/13/14 JRL
 ============================================================================}
procedure TvvaWSC.CompressPDF(UncompressedFilename, CompressedFilename: String);
var
  SEInfo: TShellExecuteInfo;
  ExitCode: DWORD;
  ExecuteFile, ParamString, StartInString : string;
begin
  ParamString := '--linearize "' + UncompressedFilename + '" "' + CompressedFilename + '"';
  ExecuteFile:='qpdf.exe';
  FillChar(SEInfo, SizeOf(SEInfo), 0) ;
  SEInfo.cbSize := SizeOf(TShellExecuteInfo) ;
  with SEInfo do 
  begin
    fMask := SEE_MASK_NOCLOSEPROCESS;
    Wnd := Application.Handle;
    lpFile := PChar(ExecuteFile) ;
    lpParameters := PChar(ParamString) ; 
    StartInString := ExtractFilePath(Application.ExeName);      
    lpDirectory := PChar(StartInString);
    nShow := SW_HIDE;
  end;

  if ShellExecuteEx(@SEInfo) then 
  begin
    repeat
      Application.ProcessMessages;
      GetExitCodeProcess(SEInfo.hProcess, ExitCode) ;
    until (ExitCode <> STILL_ACTIVE) or
    Application.Terminated;
    end;
end;



{=======================================================================
 RaveBeforePrint
 This procedure is the Rave Reports OnBeforePrint handler that provides
 a setup location for properties prior to the OnPrint event.
 =======================================================================}
procedure TVVAWSC.RaveBeforePrint(Sender: TObject);
begin
  with Sender as TBaseReport do
  begin
    LineHeight := 0.147;  //just to match existing QuickReports spacing
  end;
end;

{==================================================================
 RavePrint
 This procedure is the Rave Reports OnPrint handler that breaks the
 report text into pages.
 ==================================================================}
procedure TVVAWSC.RavePrint(Sender: TObject);
var
  I: Integer;
  currentLineHeight: Double;
  footerLineHeight: Double;
begin
  with Sender as TBaseReport do begin
    footerLineHeight := 0.2;
    SetFont(vvaReport.FontName, vvaReport.FontSize);
    for I := 0 to vvaReport.ReportText.Count - 1 do
    begin
      PrintLine(TBaseReport(Sender), vvaReport.ReportText[I]);
      currentLineHeight := LineHeight;
      LineHeight := footerLineHeight;
      if (LinesLeft < 5) then  //need 4 lines for footer
        NewPage;
      LineHeight := currentLineHeight;
    end;
  end;
end;

{============================================================================
 PrintLine
 This procedure provides a wrapper around the TBaseReport PrintLn procedure
 to prevent Tab (#9) characters from crashing the report writer.
 ============================================================================}
procedure TVVAWSC.PrintLine(ReportComponent: TBaseReport;
  ReportLine: String);
var
  cleanLine, trimmedLine: String;
  printWidth : Double;
  maxChars: Integer;
begin
  if Pos(#9, ReportLine) > 0 then
    cleanLine := ReplaceTabs(ReportLine, 6)
  else
    cleanLine := ReportLine;

  {Some reports erroneously return long strings of periods '.' that
   are intended as a keepalive for interactive terminal sessions.
   These strings cause errors when they are longer than 255 characters.
   The following code attempts to prevent lines of periods from printing and
   also wraps the remaining valid text that is longer than the available
   page print width.}
  printWidth := ReportComponent.PageWidth - ReportComponent.MarginLeft - ReportComponent.MarginRight;
  if ReportComponent.TextWidth(cleanLine) > printWidth then
  begin
    maxChars := Round(Length(cleanLine)/ReportComponent.TextWidth(cleanLine) * printWidth);
    while (Length(cleanLine) > maxChars) do
    begin
      trimmedLine := Copy(cleanLine, 0, maxChars);
      if trimmedLine <> StringOfChar('.', Length(trimmedLine)) then  //only print if not periods
        ReportComponent.PrintLn(trimmedLine);

      cleanLine := Copy(cleanLine, maxChars + 1, Length(cleanLine));
      if cleanLine = StringOfChar('.', Length(cleanLine)) then   //clear "periods only" line
        cleanLine := '';
    end;
  end;

  ReportComponent.Println(cleanLine);

end;

{==========================================================================
 ReplaceTabs
 This utility function supports the PrintLine procedure and replaces all
 tab characters with spaces in a given string based on the given tab size.
 ==========================================================================}
function TVVAWSC.ReplaceTabs(LineWithTabs: String;
  tabSize: Integer): String;
var
  I, padLength : Integer;
  LineWithSpaces : String;
begin
  LineWithSpaces := '';
  for I := 1 to Length(LineWithTabs) do
  begin
    if LineWithTabs[i] = #9 then
    begin
      padLength := tabSize - (Length(LineWithSpaces) MOD tabSize);
      LineWithSpaces := LineWithSpaces + StringOfChar(' ', padLength);
    end
    else
      LineWithSpaces := LineWithSpaces + LineWithTabs[i];
  end;
  Result := LineWithSpaces;
end;

{===================================================================
 RavePrintFooter
 This procedure is the Rave Reports OnPRintFooter handler that
 formats the report footer and occurs after each NewPage call.
 ===================================================================}
procedure TVVAWSC.RavePrintFooter(Sender: TObject);
var
  X, Y: double;
  SavedPen, LinePen: TPen;
begin
  with Sender as TBaseReport do begin
    SaveFont(1);
    SetFont('Arial', 10);
    SectionBottom := SectionBottom - (3 * LineHeight);
    GoToFooter;

    //Draw separator line
    SavedPen := Canvas.Pen;
    LinePen := CreatePen(clBlack, psSolid, 10, pmBlack);
    try
      Canvas.Pen := LinePen;
      X := XD2I(CursorXPos);
      Y := YD2I(CursorYPos);
      MoveTo(X, Y);
      LineTo(SectionRight, Y);
    finally
      Canvas.Pen := SavedPen;
      FreeAndNil(LinePen);
    end;

    //footer line 1
    SectionBottom := SectionBottom + LineHeight;
    PrintFooter(vvaReport.FooterDesc, pjLeft);
    PrintFooter('Page: ' + IntToStr(CurrentPage), pjRight);

    //footer line 2
    SectionBottom := SectionBottom + LineHeight;
    PrintFooter(vvaReport.Patient.FullName + '  SSN#' + vvaReport.Patient.SSN, pjLeft);
    PrintFooter('Printed on: ' + vvaReport.PrintDate, pjRight);

    //footer line 3
    SectionBottom := SectionBottom + LineHeight;
    PrintFooter('System: ' +  vvaReport.SystemName, pjLeft);
    PrintFooter('Division: ' + vvaReport.Division, pjRight);

    RestoreFont(1);
  end;
end;

{=====================================================================
 RavePrintHeader
 This procedure is the placeholder for the Rave Reports OnPrintHeader
 handler.  Not used at this time.
 =====================================================================}
procedure TVVAWSC.RavePrintHeader(Sender: TObject);
begin
  //Not currently used - placeholder only.
end;

{=====================================================================
 MakeUniqueFileName
 CodeCR353 LMS 2013-01-29 - added function
 CodeCR353 LMS 2013-01-13 - updated function to increase range, refactor and add SSN to unique name hash creation.

 * The intent is to create a unique string but not too long since the filename is
   limited to 32 characters.
 * Take the given file name, strip out the path and extension.
 * Note original length and Truncate to e.g. 25 chars allowing a min of 3 chars padding.
   Since length of 32 includes .pdf (4 chars)
 * Now we know how many chars we can append upto a max of 10 chars.
 * Get a hash of a GUID. Cardinal (max value 4294967295). Any seed hash less than
   (36^(CharsToPad-1)+1) is too small.
 * Use a set of A..Z, 0..9, a..z as alphanumeric set for valid characters.
 * Divide hash by size of alphanumeric set giving one char as remainder and dividend
   as the input to next value.
 * get max number of chars needed repeating the process if needed to get a second hash
   because a single cardinal is only good for giving 5 characters in a full alphanumeric
   set and 6 chars in a case-insensitive alphanumeric set.
 * Append extension to truncated name above and return.
 =====================================================================}
function TvvaWSC.makeUniqueFileName(const FileName: String; const SSN: String): String;     // CodeCR353 LMS 2013-01-29  2013-01-31
const
  AlphaNumSetLen = 36;
  // VVA filenames are case-insensitive AlphaNums : array[0..AlphaNumSetLen] of Char = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
  AlphaNums : array[0..AlphaNumSetLen] of Char = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
  MaxCharsPerHash = 6; // calculated as   iterate on MaxCharsPerHash from 1 until AlphaNumSetLen^MaxCharsPerHash > maxcardinalvalue
  MinUniqueChars = 10;
  MaxUniqueChars = 25;
  MaxFileNameLength = 32;

  function getUniqueHashString(const HashLengthNeeded: Smallint; const SSN: string) : cardinal;
  var
    baseStr: string;
    Guid: TGUID;
    // res: HRESULT;
  begin
    repeat
      // res := sysutils.CreateGUID(Guid);
      baseStr := GetStringGUID {ToString(Guid)} + SSN + IntToStr(Random(maxint));  // It appeared that repeated calls to CreateGUID returned same value twice due to a bug in code. Random is added here to circumvent similar problems. Its overkill.
      result := hash32(PChar(baseStr), length(baseStr));         // If dependance of syncommons.pas needs to be removed then consider copying not public function code for GetHashCode from graphics.pas
    until (result > (IntPower(AlphaNumSetLen, (HashLengthNeeded - 1)) + 1));
  end;

  function cvtHashToString(const aHash: cardinal; const maxStrLen: smallint): string;
  begin
    result := AlphaNums[aHash mod AlphaNumSetLen];
    if (maxStrLen > 1) then
    begin
      result := result + cvtHashToString( aHash div AlphaNumSetLen, maxStrLen - 1);
    end;
  end;

var
  aHash: cardinal;
  baseFile, FileExt: string;
  CharsToPad: SmallInt;
  CharsToTrim: smallint;
begin
  baseFile := ExtractFileName(FileName);
  FileExt := ExtractFileExt(FileName);

    // we will append at least 10 chars and at most 25 chars

  CharsToPad := min( max(MaxFileNameLength - Length(baseFile) , MinUniqueChars), MaxUniqueChars);
  CharsToTrim := Length(baseFile) - MaxFileNameLength + MinUniqueChars;

  baseFile := copy(baseFile, 0, length(baseFile) - (length(FileExt)));
  if CharsToTrim > 0 then
  begin
    baseFile := copy(baseFile, 0, length(baseFile) - CharsToTrim);
  end;

  aHash := getUniqueHashString(min(MaxCharsPerHash , CharsToPad), SSN);  // Each guid to cardinal to hash is good for only returning MaxCharsPerHash chars from alphanum set
  result := baseFile + cvtHashToString(aHash, min(MaxCharsPerHash, CharsToPad));
  while CharsToPad > MaxCharsPerHash do begin
    CharsToPad := CharsToPad - MaxCharsPerHash;
    aHash := getUniqueHashString(min(MaxCharsPerHash , CharsToPad), SSN);
    result := result + cvtHashToString(aHash, min(MaxCharsPerHash, CharsToPad));
  end;
  result := result + FileExt;
end;

{==========================================================================
 XmitFile
 This function calls the Virtual VA web service after encoding the file.
 Input:  fileName: String - path/name of local file to Xmit
 Output: function result - returns True on successful transmission;
                           otherwise, returns False
 ==========================================================================}
function TVVAWSC.XmitFile(vvaReport: TvvaReport): String;
var
  wsClaimNum, wsFileName, wsFilePath, wsDocTitle: WideString;
  wsDocSubj, wsSource, wsSourceCmmnt, wsUser, wsURL: WideString;
  wsToken, wsSecUsr, wsSecPwd: WideString;
  wsDocType, wsDocCat: Integer;
  wsUniqueFileName: widestring;  // CodeCR353 LMS 2013-01-29

  PreviousRPCCallinProgressState : Boolean; // LMS for the timer stuff and AppProxy see comments in frmVVAGetDocs
  PreviousTimeoutTimerState : Boolean;
  PreviousRPCTimeLimit: integer;

  FApplicationProxy : TApplicationProxy;
begin
  try
    wsClaimNum := vvaReport.Patient.SSN;

    wsFileName := StringReplace(vvaReport.VVADoc.OutputFileName, '\', '\\', [rfReplaceAll]);

    wsFilePath := StringReplace(vvaReport.VVADoc.OutputFilePath, '\', '\\', [rfReplaceAll])
                + wsFileName;

    // CodeCR353: LMS - Additional Requirement to make a unique file name for each send so my chance it doesn't clash with
    //  another exact same filename send by another user within the same millisecond. It happens due to file names
    //  for temporary file being same at each workstation.

    wsUniqueFileName := makeUniqueFileName(ExtractFileName(wsFileName),wsClaimNum);      // CodeCR3535 LMS 2013-01-29 and 2013-01-13

    wsDocTitle := vvaReport.FooterDesc;
    wsDocType := vvaReport.VVADoc.DocType;
    wsDocSubj := vvaReport.VVADoc.Subject;
    wsDocCat := vvaReport.vvaDoc.DocCategory;
    wsSource := vvaReport.VVADoc.Source;
    wsUser := vvaReport.VVADoc.User;
    wsSourceCmmnt := vvaReport.VVADoc.SourceComment;
    wsURL := vvaReport.VVADoc.URL;
    wsSecUsr := Piece(vvaReport.VVADoc.SecToken,'^',1);
    wsSecPwd := Piece(vvaReport.VVADoc.SecToken,'^',2);
    wsToken := Piece(vvaReport.VVADoc.SecToken,'^',3);

    // CodeCR424 LMS 2013-05-07
    // LMS for the timer stuff and AppProxy see comments in frmVVAGetDocs
    PreviousTimeoutTimerState := True;
    // PreviousRPCCallinProgressState := false;
    try
      PreviousTimeoutTimerState := frmMain.TimeoutTimer.Enabled;
      // PreviousRPCCallinProgressState := ANURemoteProcedureCallinProgress;   See Notes in frmVVAGetDocs
      frmMain.TimeoutTimer.Enabled := True;
      // ANURemoteProcedureCallinProgress := False;

      PreviousRPCTimeLimit := frmMain.RPCBroker1.RPCTimeLimit;
      frmMain.RPCBroker1.RPCTimeLimit := (PreviousRPCTimeLimit*2000);

      // FApplicationProxy := TApplicationProxy.create(); To use only later if we find that SendVVADoc causes timeouts for other CAPRI timers.
      VVAOpInProcess := True;
      try
        Result := String(SendVVADoc(wsClaimNum, wsFilePath, wsDocTitle, wsUniqueFileName,         // CodeCR3535 LMS 2013-01-29
                                    wsDocType, wsDocSubj, wsDocCat, wsSource,
                                    wsUser, wsSourceCmmnt, wsURL,
                                    wsSecUsr, wsSecPwd, wsToken));
      finally
        VVAOpInProcess := False;
        //FreeAndNil(FApplicationProxy);
        frmMain.RPCBroker1.RPCTimeLimit := PreviousRPCTimeLimit;
      end;

      frmMain.IsRPCBrokerConnected();

      Application.ProcessMessages;
    finally
      // ANURemoteProcedureCallinProgress := PreviousRPCCallinProgressState;
      frmMain.TimeoutTimer.Enabled := PreviousTimeoutTimerState;
    end;

    if (StrToIntDef(Result, 0) > 0) then
    begin
      DoUpdateStatus(actSend, stSuccess);
      Result := 'Completed transfer';
    end
    else
    begin
      DoUpdateStatus(actSend, stFailed);
    end;
  except
    on E : exception do
    begin
      ShowMessage(E.Message);
    end;  { End Exception }
  end;  { End Try/Except Block }
end;

(*  // CodeCR496 - LMS - 2014-07-14 consolidate/refactor all GetTempDir into one for Win 7 specific handling
{========================================================================
 GetTempDirectory    --  LMS - 2014-07-14 refactor GetTempDir to one method call
 This function uses the GetTempPath Windows API to determine the temp
 directory for the local workstation.  This location serves as the
 storage location for the document to be transmitted to Virtual VA.

 Input: none
 Output: TEMP directory path
 ========================================================================}
function TVVAWSC.GetTempDirectory: String;
var
  tempFolder: array[0..MAX_PATH] of Char;
begin
  GetTempPath(MAX_PATH, @tempFolder);
  result := StrPas(tempFolder);
begin
  result := GetTempDir;     // DONE: CodeCR496 - LMS - 2014-07-14 - should be commented completely and all calls directly changed to use GetTempDir
end;
*)

{===========================================================================
 GetFileName
 This function processes a string to replace all non-alphanumeric filename
 characters.  The resulting name is truncated to 28 characters to comply
 with the web service requirement of a maximum of 32 characters including
 the file suffix.

 Input:  FileDescription
 Output:  returns formatted filename after replacing any character that is
          not alphanumeric with an underscore ('_') on success; otherwise,
          returns an empty string.
 ===========================================================================}
function TVVAWSC.GetFileName(const FileDescription: String): String;
const
  VALID_CHARACTERS : set of char = ['A'..'Z','a'..'z','0'..'9'];
  REPLACEMENT_CHAR = '_';
  MAX_NAME_LENGTH = 28;
var
  c : char;
  fileName: String;
begin
  fileName := '';
  for c in FileDescription do
  begin
    if (c in VALID_CHARACTERS) then
      fileName := fileName + c
    else //replace non-alphanumeric characters
      fileName := fileName + REPLACEMENT_CHAR;
  end;
  result := Copy(fileName, 1, MAX_NAME_LENGTH);
end;

//{============================================================================
// CreateXmitFile
// This function utilizes the Synopse framework to create a PDF file directly
// from the QuickReport document content and returns the created file size
// in bytes.
//
// Input:  aQuickRpt: TQuickRep - Quick Report report object
//         afooterDesc:  String - footer description used in report footer
// Output: function result - returns file size in bytes
// ============================================================================}
//function TvvaWSC.CreateXmitFile(aQuickRpt: TQuickRep; afooterDesc: String): LongInt;
//const
//  FILE_EXTENT = '.PDF';
//begin
//  //Specify filepath
//  vvaReport.VVADoc.OutputFilePath := GetTempDirectory;
//
//  //Specify filename
//  vvaReport.VVADoc.OutputFileName := GetFileName(afooterDesc) + FILE_EXTENT;
//
//  //Create PDF file
//  ExportAsPDF(aQuickRpt, vvaReport.VVADoc.OutputFileName, vvaReport.VVADoc.OutputFilePath);
//
//  //Return file size
//  Result := GetFileSize(vvaReport.VVADoc.OutputFilePath + vvaReport.VVADoc.OutputFileName);
//end;

{=========================================================================
 GetFileSize
 This function determines the size of a given file and returns the size
 in bytes.
 =========================================================================}
function TvvaWSC.GetFileSize(const FileName: String): LongInt;
var
  aFile: File of Byte;
begin
  AssignFile(aFile, FileName);
  Reset(aFile);
  try
    Result := FileSize(aFile);
  finally
    Close(aFile);
  end;
end;

{==========================================================================
 FormatByteSize
 This function converts a byte value into a string that represents the
 number expressed as a size value in bytes, kilobytes, megabytes, or
 gigabytes, depending on the size.
 ==========================================================================}
function TvvaWSC.FormatByteSize(const bytes: Integer): string;
const
  B = 1; //byte
  KB = 1024 * B; //kilobyte
  MB = 1024 * KB; //megabyte
  GB = 1024 * MB; //gigabyte
begin
  if bytes > GB then
    result := FormatFloat('#.# GB', bytes / GB)
  else
    if bytes > MB then
      result := FormatFloat('#.# MB', bytes / MB)
    else
      if bytes > KB then
        result := FormatFloat('#.# KB', bytes / KB)
      else
        result := FormatFloat('#.# bytes', bytes) ;
end;

{ TApplicationProxy }

constructor TApplicationProxy.Create;
begin
  FWasMultiThreaded := System.IsMultiThread;
  System.IsMultiThread := True;      // LMS 2013-05-15 Since we're calling the external DLL and we need to let the memory manager know to do thread safe handling of stack.
  FProxy := Application;
  FProxyProcessMessagesPointer := Self.MethodAddress('ProcessMessages');
end;

destructor TApplicationProxy.Destroy;
begin
  //
  inherited;
  System.IsMultiThread := FWasMultiThreaded;      // LMS 2013-05-15 Reset back. No need to remain in that mode if something else doesn't need it.
end;

procedure TApplicationProxy.ProcessMessages;
begin
  TApplication(FProxy).ProcessMessages;
  // inherited;
  // just to call the base class's method.
end;

end.
